part of fields.model;

/// A reflected instance variable or getter of an [Object].
/// 
/// The type parameter `T` documents the expected type of the underlying 
/// object.  It is asserted to be assignable from the actual reflected type of
/// the underlying object in checked mode only.
class Field<T> {
  
  /// Creates a new field for the instance getter with the given [name] in the 
  /// given [owner] object.
  /// 
  /// An optional [getter] mirror and [parent] field may also be given.  In
  /// general these need not be specified; They are exposed as optional 
  /// parameters to allow editor implementations to optimize by caching and 
  /// reusing these values for multiple fields.
  factory Field.forGetter(
      InstanceMirror owner, 
      Symbol name, 
     {MethodMirror getter,
      Field parent}) {
    if (getter == null) {
      getter = owner.type.declarations[name];
      if (getter == null || 
          getter is! MethodMirror || 
          !getter.isGetter || 
          getter.isStatic) {
        throw new ArgumentError(
            '$name is not an instance getter of ${owner.type.qualifiedName}');
      }
    }
    final setter = new Symbol('${MirrorSystem.getName(name)}=');
    final metadata = _parseMetadata(getter);      
    if (metadata[IgnoreField]) {
      return null;
    }   
    return new Field._(
        owner, 
        name, 
        metadata[RuntimeType] != null 
        ? metadata[RuntimeType] 
        : _typeOf(getter.returnType),
        parent: parent,
        editor: metadata[UseFieldEditor],
        transform: metadata[UseFieldTransform],
        annotations: metadata[FieldAnnotation],
        displayName: metadata[DisplayName],
        isPrivate: getter.isPrivate,
        isReadOnly: !owner.type.declarations.containsKey(setter) 
          || metadata[ReadOnly],
        isVariable: false);
  }
  
  /// Creates a new field for the instance variable with the given [name] in 
  /// the given [owner] object.
  /// 
  /// An optional [variable] mirror and [parent] field may also be given.  In
  /// general these need not be specified; They are exposed as optional 
  /// parameters to allow editor implementations to optimize by caching and 
  /// reusing these values for multiple fields.
  factory Field.forVariable(
      InstanceMirror owner, 
      Symbol name, 
     {VariableMirror variable,
      Field parent}) {
    if (variable == null) {
      variable = owner.type.declarations[name];
      if (variable == null || 
          variable is! VariableMirror || 
          variable.isStatic) {
        throw new ArgumentError(
            '$name is not an instance variable of ${owner.type.qualifiedName}');
      }
    }
    final metadata = _parseMetadata(variable);      
    if (metadata[IgnoreField]) {
      return null;
    }
    return new Field._(
        owner, 
        name, 
        metadata[RuntimeType] != null 
        ? metadata[RuntimeType] 
        : _typeOf(variable.type),
        parent: parent,
        editor: metadata[UseFieldEditor],
        transform: metadata[UseFieldTransform],
        annotations: metadata[FieldAnnotation],
        displayName: metadata[DisplayName],
        isPrivate: variable.isPrivate,
        isReadOnly: variable.isFinal || metadata[ReadOnly]);
  }
  
  /// The name of this field as declared in the [owner]'s class declaration.
  final Symbol name;
  
  /// The display name of this field.
  /// 
  /// By default this is equal to the [name] of this field.  It may be overriden
  /// by annotating the field with a [DisplayName] annotation.
  final String displayName;
  
  /// The type of this field's [value].
  final Type type;
  
  /// Returns `true` if this field is declared as a getter.
  bool get isGetter => !isVariable;
  
  /// Returns `true` if this field is declared as private.
  final bool isPrivate;
  
  /// Returns `true` if this field is read-only.
  /// 
  /// A field is read-only if it reflects a `final` variable, a getter with no
  /// matching setter, or if it is annotated with [ReadOnly].
  final bool isReadOnly;
  
  /// Returns `true` if this field is declared as an instance variable.
  final bool isVariable;
  
  /// The parent of this field or `null` if this field has no parent.
  final Field parent;
  
  /// A mirror on the class of this field's editor; this is `null` by default.
  /// 
  /// By default, this is `null` and the choice of a field editor for this field
  /// is typically based on the [type].  If the field is annotated with 
  /// [UseFieldEditor] then this will be the reflected class of the annotated
  /// editor type.
  final ClassMirror editor;
  
  /// This field's preferred transform; this is `null` by default.
  /// 
  /// By default, this is `null` and the choice of a field transform for this
  /// field is chosen by an editor based on the [type].  If the field is
  /// annotated with [UseFieldTransform] then this will be the annotated 
  /// transform.
  final FieldTransform transform;
  
  /// Any field annotations that have been applied to this field outside of the
  /// set of field annotations defined in this library.
  /// 
  /// The set of field annotations defined in this library ([DisplayName], 
  /// [ReadOnly], etc...) have a direct effect on one or more properties of this
  /// field and are therefore not included here.
  /// 
  /// This collection provides a mechanism for annotating fields with custom
  /// metadata that may be accessed later by custom editors.
  final Iterable<FieldAnnotation> annotations;
  
  FieldList _children;
  /// The children of this field.
  /// 
  /// A field's children does not include those whose [type] is determined to be 
  /// `simple`.  The definition of `simple` used here is more inclusive than the 
  /// term as used in `dart:mirrors` and includes `Map`, `List`, `dynamic`, 
  /// `Type`, and `Object`.
  FieldList get children => _children;
  
  /// A mirror on the object instance that 'owns' this field.
  InstanceMirror owner;  
  
  Object _value;
  /// The current value of this field.
  Object 
    get value => _value;
    set value(Object value) {
      if(isReadOnly) {
        throw new StateError('Field $displayName is read-only.');
      }
      if (value != _value) {
        owner.setField(name, value);
        _value = value;
        _refreshChildren().then((_) {
          _values.add(value);
          if (parent != null) {
            parent._values.add(parent._value);
          }
        });
      }
    }
  
  final _values = new StreamController<Object>.broadcast();
  /// An observable stream of this field's values as they change.
  Stream<Object> get values => _values.stream;
  
  Field._(
      this.owner, 
      Symbol name, 
      this.type, 
     {this.parent, 
      this.editor, 
      this.transform,
      this.annotations: const [], 
      String displayName, 
      this.isPrivate: false, 
      this.isReadOnly: false,
      this.isVariable: true})
      : this.name = name
      , this.displayName = 
          (displayName == null) ? MirrorSystem.getName(name) : displayName {
    assert(_isAssignableFrom(reflectType(T), reflectType(type)));
    _children = new FieldList(this);
    _value = owner.getField(name).reflectee;
    if (!_isSimpleType(type)) {
      _children.refreshSource(_value);
    }
  }
  
  /// Refresh this field's [value].
  /// 
  /// A field can only know about changes to its [value] that happen through
  /// this interface.  If another object changes the underlying value in some 
  /// other way then this method should be called.
  void refreshValue() {
    final value = owner.getField(name).reflectee;
    if (value != _value) {
      _value = value;
      _refreshChildren().then((_) => _values.add(value));
    }
  }
        
  Future _refreshChildren() {    
    if (children.isEmpty) {
      return new Future.value();
    }
    children.refreshSource(_value);
    return Future.forEach(children, (f) => new Future(f.refreshValue));
  }
}
