part of fields.model;

Map _parseMetadata(DeclarationMirror declaration) {
  assert(declaration != null);
  final metadata = new Map()        
  ..[UseFieldEditor] = null
  ..[UseFieldTransform] = null
  ..[ReadOnly] = false
  ..[IgnoreField] = false
  ..[DisplayName] = null
  ..[RuntimeType] = null
  ..[FieldAnnotation] = <FieldAnnotation>[];
  declaration.metadata.forEach((InstanceMirror mirror) {
    final instance = mirror.reflectee;
    if (instance is UseFieldEditor) {
      // Disallow > 1 @UseFieldEditor
      assert(metadata[UseFieldEditor] == null);
      // TODO(rms): validate that the editor type isAssignableTo FieldEditor.
      metadata[UseFieldEditor] = reflectClass(instance.editorType);
    } else if (instance is UseFieldTransform) {
      // Disallow > 1 @UseFieldTransform
      assert(metadata[UseFieldTransform] == null);
      metadata[UseFieldTransform] = instance.transform;
    } else if (instance == ReadOnly) {
      metadata[ReadOnly] = true;
    } else if (instance == IgnoreField) {
      metadata[IgnoreField] = true;
    } else if (instance is DisplayName) {
      // Disallow > 1 @DisplayName
      assert(metadata[DisplayName] == null);
      metadata[DisplayName] = instance.name;
    } else if (instance is RuntimeType) {
      // Disallow > 1 @RuntimeType
      assert(metadata[RuntimeType] == null);
      metadata[RuntimeType] = instance.type;
    } else if (instance is FieldAnnotation) {
      // Editor-specific annotations, store them in the field as-is.
      metadata[FieldAnnotation].add(instance);
    }
  });
  return metadata;
}

/// An annotation that may be applied to an instance variable or getter of an 
/// object's class declaration.
/// 
/// A default set of field annotations are defined in this library.  This
/// annotation may be extended or implemented to provide metadata to custom
/// field editors.  Custom field annotations are made available to editors as
/// [Field.annotations].
abstract class FieldAnnotation {
  const FieldAnnotation();  
}

/// An annotation that declares a field should be ignored by editors.
///
/// A call to [new Field.forVariable] or [new Field.forGetter] will return 
/// `null` if this annotation is found in the field's declaration.
const FieldAnnotation IgnoreField = const _IgnoreField();
class _IgnoreField extends FieldAnnotation {
  const _IgnoreField();
}

/// An annotation that declares the display name for a field.
///
/// By default the display name of a field is the name of the variable or getter
/// as given by [Field.name].
class DisplayName extends FieldAnnotation {
  
  /// The display name for the annotated field.
  final String name;
  
  /// Create a constant display name annotation with the given [name].
  const DisplayName(this.name);
}

/// An annotation that declares a field to be read-only.
/// 
/// The default value of [Field.isReadOnly] is `true` for final variables or a
/// getter with no matching setter.  This annotation provides an override to 
/// declare that a field be considered read-only if the default value is 
/// `false`.
const FieldAnnotation ReadOnly = const _ReadOnly();
class _ReadOnly extends FieldAnnotation {
  const _ReadOnly();
}

/// This is a workaround for the following issues, and will be removed as soon
/// as they are resolved:
/// 
/// * <http://dartbug.com/6433>
/// * <http://dartbug.com/12022>
class RuntimeType extends FieldAnnotation {
  final Type type;
  const RuntimeType(this.type);  
}

/// An annotation that declares the use of a specific [FieldEditor] for a field.
/// 
/// By default a field editor is chosen based on a field's `runtimeType`.  This
/// annotation provides an override to declare that a field use an instance of
/// the specified [editorType].
class UseFieldEditor extends FieldAnnotation {
  
  /// The editor type to use for the annotated field.
  final Type editorType;
  
  /// Create a constant field editor annotation with the given [editorType].
  /// 
  /// The given [editorType] must extend or implement [FieldEditor].
  const UseFieldEditor(this.editorType);
}

/// An annotation that declares the use of a specific [FieldTransform] for a 
/// field.
/// 
/// By default a field transform is chosen based on a field's `runtimeType`.  
/// This annotation provides an override to declare that a field use the given
/// [transform] instead.
class UseFieldTransform extends FieldAnnotation {
  
  /// The transform to use for the annotated field.
  final FieldTransform transform;  
  
  /// Create a constant field transform annotation with the given [transform].
  const UseFieldTransform(this.transform);
}

class UseRange extends FieldAnnotation {
  
  final num min, max, step;
  const UseRange({this.min, this.max, this.step});
  
  /// Applies this range to the given [object].
  /// 
  /// The [object] is expected to have fields named `min`, `max`, and `step`;
  /// an example of such an object is an instance of [InputElement].
  void applyTo(object) {
    if (min != null) object.min = min.toString();
    if (max != null) object.max = max.toString();      
    if (step != null) object.step = step.toString();
  }
}

class UseValues extends FieldAnnotation {
  final List values;
  const UseValues(this.values);
}
