part of fields.view;

class BoolEditor extends FieldEditor<bool> {
  CheckboxInputElement _input;
  Element get element => _input;
  
  BoolEditor(Field field) : super(field) {
    _input = input(() {
      style()
      ..scoped = true
      ..text = resources.style; 
    })
    ..type = 'checkbox'
    ..onChange.listen((_) {
      this.value = _input.checked;
    });
    
    render();
  }
      
  void render() {
    _input
    ..disabled = isValueReadOnly
    ..checked = value;
  }
}

class ColorEditorResources extends FieldEditorResources {
  String get style => super.style + 
"""
.picker {  
  float:left;
}
.valuebox { 
  overflow:hidden; 
}
""";
  const ColorEditorResources();
}

class ColorEditor extends FieldEditor<String> {
  InputElement _picker;
  InputElement _display;
  Element _element;
  Element get element => _element;
  
  ColorEditor(Field field) : super(field, const ColorEditorResources()) {
    _element = span(() {
      style()
      ..scoped = true
      ..text = resources.style;   
      _picker = input()
      ..type = 'color'
      ..classes = ['picker']
      ..onChange.listen((_) {
        _display.value = _picker.value;         
        this.value = _picker.value;
      });
      div(() {
        _display = input()
        ..type = 'text'
        ..disabled = true;
      })
      ..classes=['valuebox'];
    });
    
    render();
  }
  
  void render() {
    _picker
    ..disabled = isValueReadOnly
    ..value = value;
    _display.value = value;
  }
}

class DateTimeEditor extends FieldEditor<DateTime> {
  DateInputElement _input;
  Element get element => _input;
  
  DateTimeEditor(Field field) : super(field) {
    _input = input(() {
      style()
      ..scoped = true
      ..text = resources.style; 
    })
    ..type = 'date' // TODO(rms): investigate 'datetime-local'
    ..onChange.listen((_) {
      this.value = _input.valueAsDate;
    });
    
    render();
  }
  
  void render() {
    _input
    ..disabled = isValueReadOnly
    ..valueAsDate = value;
  }
}

class DefaultEditor extends FieldEditor<String> {
  InputElement _input;
  Element get element => _input;
    
  DefaultEditor(Field field) : super(field) {
    _input = input(() {
      style()
      ..scoped = true
      ..text = resources.style; 
    })
    ..type = 'text'
    ..onInput.listen((_) {
      try {
        this.value = _input.value;
        _input.classes.remove('error');
      } catch(_) {
        _input.classes.add('error');
      }
    });
    
    render();
  }
  
  void render() {
    _input
    ..disabled = isValueReadOnly
    ..value = value;
  }
}

class NumberEditor extends FieldEditor<num> {
  InputElement _input;
  Element get element => _input;
  
  NumberEditor(Field field) : super(field) {
    _input = input(() {
      style()
      ..scoped = true
      ..text = resources.style; 
    })
    ..type = 'number'
    ..onChange.listen((_) {
      this.value = _input.valueAsNumber;
    });
    
    final match = (a) => a is UseRange;
    if (field.annotations.any(match)) {
      UseRange range = field.annotations.firstWhere(match);
      range.applyTo(_input);
    }
    
    render();
  }
  
  void render() {
    _input
    ..disabled = isValueReadOnly
    ..valueAsNumber = value;
  }
}

class RangeEditorResources extends FieldEditorResources {
  String get style => super.style + 
"""
.range {  
  float:left;
}
.valuebox { 
  overflow:hidden; 
}
""";
  const RangeEditorResources();
}

class RangeEditor extends FieldEditor<num> {
  InputElement _range;
  InputElement _display;
  Element _element;
  Element get element => _element;
  
  RangeEditor(Field field) : super(field, const RangeEditorResources()) {
    _element = span(() {
      style()
      ..scoped = true
      ..text = resources.style;   
      
      _range = input()
      ..type = 'range'
      ..classes = ['range']
      ..onChange.listen((_) {
        _display.value = _range.value; 
        this.value = _range.valueAsNumber;
      });
      
      div(() {
        _display = input()
        ..type = 'text'
        ..disabled = true;
      })
      ..classes=['valuebox'];
    });

    final match = (a) => a is UseRange;
    if (field.annotations.any(match)) {
      UseRange range = field.annotations.firstWhere(match);
      range.applyTo(_range);
    }
    
    render();
  }
  
  void render() {
    _range
    ..disabled = isValueReadOnly
    ..valueAsNumber = value;
    _display.value = value.toString();
  }
}

class SelectEditorResources extends FieldEditorResources {
  String get style => 
"""
select {
  background-color:transparent;
  border: 0;
  border-radius: 0;
  color:${fieldsTheme.colors.text};
}
option {
  background: ${fieldsTheme.colors.fieldBackground};
}
""";
  const SelectEditorResources();
}

class SelectEditor extends FieldEditor<String> {
  final List _optionValues = new List();
  SelectElement _select;
  Element get element => _select;
      
  SelectEditor(Field field) : super(field, const SelectEditorResources()) {
    _select = select(() {
      style()
      ..scoped = true
      ..text = resources.style; 
      
      final isUseValues = (a) => a is UseValues;
      if (field.annotations.any(isUseValues)) {
        UseValues useValues = field.annotations.firstWhere(isUseValues);
        _optionValues.addAll(useValues.values);        
        for (var o in _optionValues) {
          String valueString = transformField(o, o.runtimeType, String);
          option()
          ..text = valueString
          ..value = valueString;
        }
      }
    })
    ..onChange.listen((_) {
      this.value = _select.value;
    });
    
    render();
  }
  
  void render() {
    _select
    ..disabled = isValueReadOnly
    ..value = value;
  }
}
