@TestGroup(description: 'FieldEditor')
library fields.test.model.field_editor;

import 'dart:mirrors';
import 'package:bench/bench.dart';
import 'package:unittest/unittest.dart';
import 'package:fields/src/model.dart';

main() => reflectTests();

// -----------------------------------------------------------------------------
// Setup
// -----------------------------------------------------------------------------

@Setup
void setup() {
  FieldEditor.unregisterAll();
  registerDefaultFieldTransforms();
}

class Rectangle {
  int x, y, w, h;
  Rectangle(this.x, this.y, this.w, this.h);
  String toString() => 'rectangle($x,$y,$w,$h)';
}

class HasFields {
  int i = 7;
  String s = 'snarf';
  Rectangle r = new Rectangle(1, 2, 3, 4);
}

class NotAnEditor { }

class IntEditor extends FieldEditor<int> {
  final element = null;
  IntEditor(Field field) : super(field);
  void render() {}
}

class StringEditor extends FieldEditor<String> {
  final element = null;
  StringEditor(Field field) : super(field);
  void render() {}
}

class RectangleEditorResources extends FieldEditorResources {
  String get style => 'SOME CSS';
  const RectangleEditorResources();
}

class RectangleEditor extends FieldEditor<Rectangle> {
  final element = null;
  RectangleEditor(Field field) : super(field, const RectangleEditorResources());
  void render() {}
}

// -----------------------------------------------------------------------------
// Register
// -----------------------------------------------------------------------------

@Test('Test that no default field editor is initially registered.')
void testHasNoDefaultFieldEditor() {
  expect(FieldEditor.isDefaultRegistered, isFalse);
}

@Test('Test that register default throws if the given editor type is null.')
@ExpectError(isArgumentError)
void testRegisterDefaultNullThrows() {  
  FieldEditor.registerDefault(null);
}

@Test('Test that register throws if the given editor type is null.')
@ExpectError(isArgumentError)
void testRegisterNullEditorTypeThrows() {
  FieldEditor.register(int, null);
}

@Test('Test the successful register of a default editor.')
void testRegisterDefault() {
  expect(FieldEditor.isDefaultRegistered, isFalse);
  FieldEditor.registerDefault(IntEditor);
  expect(FieldEditor.isDefaultRegistered, isTrue);
}

@Test('Test that a second attempt to register a default editor throws.')
@ExpectError(isStateError)
void testRegisterDefaultTwiceThrows() {
  expect(FieldEditor.isDefaultRegistered, isFalse);
  FieldEditor.registerDefault(IntEditor);  
  FieldEditor.registerDefault(IntEditor);
}

@Test('Test that an attempt to register default a non editor type  throws.')
@ExpectError(isArgumentError)
void testRegisterDefaultNonEditorTypeThrows() {
  expect(FieldEditor.isDefaultRegistered, isFalse);
  FieldEditor.registerDefault(NotAnEditor);
}

@Test('Test the successful register of an int editor.')
void testRegisterIntEditor() {
  expect(FieldEditor.isRegistered(int), isFalse);
  FieldEditor.register(int, IntEditor);
  expect(FieldEditor.isRegistered(int), isTrue);
}

@Test('Test that a second attempt to register an int editor throws.')
@ExpectError(isStateError)
void testRegisterIntEditorTwiceThrows() {
  expect(FieldEditor.isRegistered(int), isFalse);
  FieldEditor.register(int, IntEditor);  
  FieldEditor.register(int, IntEditor);
}

@Test('Test that an attempt to register a non editor type throws.')
@ExpectError(isArgumentError)
void testRegisterNonEditorTypeThrows() {
  expect(FieldEditor.isRegistered(int), isFalse);
  FieldEditor.register(int, NotAnEditor);
}

// -----------------------------------------------------------------------------
// Factory
// -----------------------------------------------------------------------------

@Test('Test that the factory throws if no field editor is registered.')
@ExpectError(isStateError)
void testFactoryNoEditorThrows() {
  final obj = new HasFields();
  final mirror = reflect(obj);
  final field = new Field.forVariable(mirror, #i);
  var editor = new FieldEditor.forField(field);
}

@Test('Test that the factory returns the default editor.')
void testFactoryReturnsDefaultEditor() {
  FieldEditor.registerDefault(IntEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final i = new Field.forVariable(mirror, #i);
  final s = new Field.forVariable(mirror, #s);
  final iEditor = new FieldEditor.forField(i);
  final sEditor = new FieldEditor.forField(s);
  expect(iEditor, new isInstanceOf<IntEditor>());
  expect(iEditor.field, equals(i));
  expect(sEditor, new isInstanceOf<IntEditor>());
  expect(sEditor.field, equals(s));
}

@Test('Test that the factory returns registered editors with no default.')
void testFactoryReturnsNoDefaultEditor() {
  FieldEditor.register(int, IntEditor);
  FieldEditor.register(String, StringEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final i = new Field.forVariable(mirror, #i);
  final s = new Field.forVariable(mirror, #s);
  final iEditor = new FieldEditor.forField(i);
  final sEditor = new FieldEditor.forField(s);
  expect(iEditor, new isInstanceOf<IntEditor>());
  expect(iEditor.field, equals(i));
  expect(sEditor, new isInstanceOf<StringEditor>());
  expect(sEditor.field, equals(s));
}

// -----------------------------------------------------------------------------
// Value
// -----------------------------------------------------------------------------

@Test('Test getting the value of an int field in an int editor.')
void testIntFieldIntEditorGetValue() {
  FieldEditor.register(int, IntEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final i = new Field.forVariable(mirror, #i);
  IntEditor editor = new FieldEditor.forField(i);
  expect(editor.isValueReadOnly, isFalse);
  expect(editor.value, equals(7));
}

@Test('Test getting the value of an int field in a string editor.')
void testIntFieldStringEditorGetValue() {
  FieldEditor.registerDefault(StringEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final i = new Field.forVariable(mirror, #i);
  StringEditor editor = new FieldEditor.forField(i);
  expect(editor.isValueReadOnly, isFalse);
  expect(editor.value, equals('7'));
}

@Test('Test getting the value of a string field in an int editor throws.')
@ExpectError(const isInstanceOf<FieldTransformError>())
void testStringFieldIntEditorGetValueThrows() {
  FieldEditor.registerDefault(IntEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final s = new Field.forVariable(mirror, #s);
  IntEditor editor = new FieldEditor.forField(s);
  // This should throw because the value 'snarf' cannot be parsed as an int.
  var v = editor.value;
}

@Test('Test that the value of a rectangle field in a string editor is '
      'read-only when no string -> rectangle transform is registered.')
void testRectagleFieldStringEditorValueReadOnly() {
  FieldEditor.registerDefault(StringEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final r = new Field.forVariable(mirror, #r);
  StringEditor editor = new FieldEditor.forField(r);
  // There is no registered String -> Rectangle transform so we expect readonly.
  expect(editor.isValueReadOnly, isTrue);
  expect(editor.value, equals('rectangle(1,2,3,4)'));
}

@Test('Test setting the value of an int field in an int editor.')
void testIntFieldIntEditorSetValue() {
  FieldEditor.register(int, IntEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final i = new Field.forVariable(mirror, #i);
  IntEditor editor = new FieldEditor.forField(i);
  expect(editor.isValueReadOnly, isFalse);
  editor.value = 42;
  expect(editor.value, equals(42));
  expect(obj.i, equals(42));
}

@Test('Test setting the value of an int field in a string editor.')
void testIntFieldStringEditorSetValue() {
  FieldEditor.registerDefault(StringEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final i = new Field.forVariable(mirror, #i);
  StringEditor editor = new FieldEditor.forField(i);
  expect(editor.isValueReadOnly, isFalse);
  editor.value = '42';
  expect(editor.value, equals('42'));
  expect(obj.i, equals(42));
}

@Test('Test setting bad value of an int field in a string editor throws.')
@ExpectError(const isInstanceOf<FieldTransformError>())
void testIntFieldStringEditorSetValueThrows() {
  FieldEditor.registerDefault(StringEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final i = new Field.forVariable(mirror, #i);
  StringEditor editor = new FieldEditor.forField(i);
  expect(editor.isValueReadOnly, isFalse);
  // This should throw because the value cannot be parsed as an int.
  editor.value = '42bottlesofbeeronthewall';
}

@Test('Test that setting the value of a rectangle field in a string '
      'editor throws because the editor is read-only.')
@ExpectError(isUnsupportedError)      
void testRectangleFieldStringEditorSetValueThrows() {
  FieldEditor.registerDefault(StringEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final r = new Field.forVariable(mirror, #r);
  StringEditor editor = new FieldEditor.forField(r);
  // There is no registered String -> Rectangle transform so we expect readonly.
  expect(editor.isValueReadOnly, isTrue);
  // Setting a value when read-only is expected to throw.
  editor.value = new Rectangle(5,6,7,8);
}

// -----------------------------------------------------------------------------
// Resources
// -----------------------------------------------------------------------------

@Test('Test that a rectangle editor may specify a custom resources type.')
void testRectangleEditorResources() {
  FieldEditor.register(Rectangle, RectangleEditor);
  final obj = new HasFields();
  final mirror = reflect(obj);
  final r = new Field.forVariable(mirror, #r);
  FieldEditor editor = new FieldEditor.forField(r);
  expect(editor.isValueReadOnly, isFalse);
  expect(editor, const isInstanceOf<RectangleEditor>());
  expect(editor.resources, const isInstanceOf<RectangleEditorResources>());
  expect(editor.resources.style, equals('SOME CSS'));
}
