@TestGroup(description: 'FieldTransform')
library fields.test.model.field_transform;

import 'package:bench/bench.dart';
import 'package:unittest/unittest.dart';
import 'package:fields/src/model.dart';

main() => reflectTests();

// -----------------------------------------------------------------------------
// Setup
// -----------------------------------------------------------------------------

@Setup
void setup() => registerDefaultFieldTransforms();

@Teardown
void teardown() => FieldTransform.unregisterAll();

class CustomStringToIntTransform extends FieldTransform<String, int> {
  static int _string2int(int value) => throw new UnimplementedError();
  const CustomStringToIntTransform() : super(_string2int);
}

class Rectangle {
  int x, y, w, h;
}

class RectangleF {
  double x, y, w, h;
}

class RectangleToRectangleFTransform 
    extends FieldTransform<Rectangle, RectangleF> {
  static RectangleF _r2rf(Rectangle value) => 
      new RectangleF()..x = value.x.toDouble()
      ..y = value.y.toDouble()
      ..w = value.w.toDouble()
      ..h = value.h.toDouble();
  const RectangleToRectangleFTransform() : super(_r2rf);
}

// -----------------------------------------------------------------------------
// Basic
// -----------------------------------------------------------------------------

@Test('Test that simple types can be transformed.')
void testCanTransformSimpleTypes() {
  expect(canTransformField(int, int), isTrue);
  expect(canTransformField(String, int), isTrue);
}

@Test('Test a valid string to int transform.')
void testTransformStringToInt() {
  final String s = '42';
  final int i = transformField(s, String, int);
  expect(i, equals(42));
}

@Test('Test an erroneous string to int transform.')
@ExpectError(const isInstanceOf<FieldTransformError>())
void testTransformStringToIntThrows() {
  final String s = '42.123';
  transformField(s, String, int);
}

@Test('Test that a string to int transform where value is! a string throws.')
@ExpectError(const isInstanceOf<TypeError>())
void testTransformStringToIntValueIsNotStringThrows() {
  final map = { 'foo' : 'bar' };
  transformField(map, String, int); 
}

@Test()
void testTransformNotHasInverse() {
  var xform = new FieldTransform<int, bool>((_) {});
  expect(xform, isNotNull);
  expect(xform.hasInverse, isFalse);
}

@Test()
void testTransformHasInverse() {
  var xform = new FieldTransform<int, bool>((_) {}, (_) {});
  expect(xform, isNotNull);
  expect(xform.hasInverse, isTrue);
}

@Test()
void testTransformInverse() {
  final letters = const [ 'a', 'b', 'c'];
  var xform = new FieldTransform<int, String>(
      (i) => letters[i], 
      (s) => letters.indexOf(s));  
  expect(xform.hasInverse, isTrue);
  expect(xform(2), equals('c'));
  expect(xform.inverse('b'), equals(1));
  expect(xform.inverse(xform(0)), equals(0));
}

@Test()
@ExpectError(isUnsupportedError)
void testTransformNoInverseThrows() {
  var xform = new FieldTransform<int, bool>((i) => i >= 1);
  xform.inverse(false);
}

// -----------------------------------------------------------------------------
// Custom
// -----------------------------------------------------------------------------

@Test('Test the registration of a custom string to int transform.')
void testRegisterCustomTransformStringToInt() {
  FieldTransform.register(new CustomStringToIntTransform());  
  expect(canTransformField(String, int), isTrue);
}

@Test('Test the registration of a custom rectangle transform.')
void testRegisterRectangleTransform() {
  expect(canTransformField(Rectangle, RectangleF), isFalse);
  final transform = new RectangleToRectangleFTransform();
  FieldTransform.register(transform);
  expect(canTransformField(Rectangle, RectangleF), isTrue);
  expect(FieldTransform.unregister(transform), isTrue);
  expect(canTransformField(Rectangle, RectangleF), isFalse);
}

@Test('Test a valid rectangle transform.')
void testTransformRectangle() {  
  final transform = new RectangleToRectangleFTransform();
  FieldTransform.register(transform);  
  final r = new Rectangle()..x = 1..y = 2..w = 3..h = 4;
  final rf = transformField(r, Rectangle, RectangleF);
  expect(rf, const isInstanceOf<RectangleF>());
  expect(rf.x, equals(1));
  expect(rf.y, equals(2));
  expect(rf.w, equals(3));
  expect(rf.h, equals(4));
  expect(FieldTransform.unregister(transform), isTrue);
}
